/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.adapter

import android.content.res.Configuration
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.LinearLayout
import androidx.constraintlayout.widget.ConstraintSet
import androidx.constraintlayout.widget.ConstraintSet.*
import androidx.recyclerview.widget.DiffUtil
import androidx.recyclerview.widget.RecyclerView
import com.kakao.sdk.partner.friend.imageloader.ImageLoader
import com.kakao.sdk.partner.friend.model.Item
import com.kakao.sdk.partner.friend.util.LayoutUtils
import com.kakao.sdk.partner.friend.util.getDefaultProfileImage
import com.kakao.sdk.v2.partner.friend.R
import com.kakao.sdk.v2.partner.friend.databinding.KakaoSdkCategoryDividerBinding
import com.kakao.sdk.v2.partner.friend.databinding.KakaoSdkEmptySearchResultTextBinding
import com.kakao.sdk.v2.partner.friend.databinding.KakaoSdkHeaderFriendPickerBinding
import com.kakao.sdk.v2.partner.friend.databinding.KakaoSdkItemFriendPickerBinding

/**
 * @suppress
 */
internal class FriendsListAdapter(
    private val isPopup: Boolean,
    private val isSingle: Boolean,
    private val enableSearch: Boolean,
    private val showPickedFriend: Boolean,
    private val maxPickableCount: Int,
    private val checkCallback: (Item.User?) -> Unit,
    private val uncheckCallback: (Item.User) -> Unit
) : RecyclerView.Adapter<RecyclerView.ViewHolder>() {
    private var items = mutableListOf<Item>() // 화면에 보여지는 전체 아이템 리스트
    private var pickedFriends = listOf<Item.User>() // 친구 추가 선택 가능 여부 파악을 위한 리스트
    private lateinit var mRecyclerView: RecyclerView // 스크롤 처리를 위한 recyclerview
    private val diffUtilCallback = ItemDiffUtilCallback()

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): RecyclerView.ViewHolder {
        val inflater = LayoutInflater.from(parent.context)
        return when (viewType) {
            VIEW_TYPE_HEADER -> {
                HeaderViewHolder(KakaoSdkHeaderFriendPickerBinding.inflate(inflater, parent, false))
            }
            VIEW_TYPE_USER -> {
                UserViewHolder(isPopup,
                    KakaoSdkItemFriendPickerBinding.inflate(inflater, parent, false),
                    checkCallback = {
                        // 최대 선택 가능한 친구를 다 채운 경우
                        if (pickedFriends.size == maxPickableCount) {
                            // 싱글 피커인 경우 이전에 선택된 친구 삭제 후 새로운 친구 리턴
                            if (isSingle) {
                                uncheckCallback(pickedFriends[0])
                                checkCallback(it)
                            } else {
                                // 멀티 피커이면서 이미 최대 선택 가능한 친구를 다 채운 경우 null 리턴
                                checkCallback(null)
                            }
                        } else {
                            checkCallback(it)
                        }
                    },
                    uncheckCallback = {
                        uncheckCallback(it)
                    })
            }
            VIEW_CATEGORY_DIVIDER -> {
                DividerViewHolder(KakaoSdkCategoryDividerBinding.inflate(inflater, parent, false))
            }
            VIEW_EMPTY_SEARCH_RESULT_TEXT -> {
                EmptySearchViewHolder(
                    isPopup,
                    KakaoSdkEmptySearchResultTextBinding.inflate(inflater, parent, false)
                )
            }
            else -> throw ClassCastException("Unknown viewType $viewType")
        }
    }

    override fun onBindViewHolder(holder: RecyclerView.ViewHolder, position: Int) {
        val item = items[position]
        when (holder) {
            is HeaderViewHolder -> holder.bind(item as Item.Header)
            is UserViewHolder -> holder.bind(item as Item.User, isSingle)
            is EmptySearchViewHolder -> holder.setLayout()
        }
    }

    override fun getItemCount(): Int = items.size

    override fun getItemViewType(position: Int): Int {
        return when (items[position]) {
            is Item.Header -> VIEW_TYPE_HEADER
            is Item.User -> VIEW_TYPE_USER
            is Item.Divider -> VIEW_CATEGORY_DIVIDER
            is Item.EmptySearchResult -> VIEW_EMPTY_SEARCH_RESULT_TEXT
        }
    }

    override fun onAttachedToRecyclerView(recyclerView: RecyclerView) {
        super.onAttachedToRecyclerView(recyclerView)
        mRecyclerView = recyclerView
    }

    fun setFriendsList(items: MutableList<Item>) {
        diffUtilCallback.setLists(this.items, items)
        DiffUtil.calculateDiff(diffUtilCallback).dispatchUpdatesTo(this)
        this.items = items
    }

    fun checkFriend(friend: Item.User) {
        if (!isSingle && !enableSearch && showPickedFriend && pickedFriends.isEmpty()) {
            items.add(0, Item.Divider)
            notifyItemInserted(0)
            // 구분선 추가 전 화면이 리스트 최상단이면 새로 추가된 구분선까지 스크롤을 위로 올려줌.
            if (mRecyclerView.computeVerticalScrollOffset() == 0) {
                mRecyclerView.scrollToPosition(0)
            }
        }
        updateDuplicatedFriend(friend, true)
    }

    fun uncheckFriend(friend: Item.User) {
        if (!isSingle && !enableSearch && showPickedFriend && pickedFriends.size == 1) {
            items.removeAt(0)
            notifyItemRemoved(0)
        }
        updateDuplicatedFriend(friend, false)
    }

    fun setPickedFriends(pickedFriends: MutableList<Item.User>) {
        this.pickedFriends = pickedFriends
    }

    private fun updateDuplicatedFriend(user: Item.User, isChecked: Boolean) {
        val indexOfFirst = items.indexOfFirst { it == user }
        val indexOfLast = items.indexOfLast { it == user }
        // index를 못찾은 경우에는 -1 이므로 crash 방지 (검색 중에는 리스트에 해당 user가 없을 수 있음)
        if (indexOfFirst > -1) {
            (items[indexOfFirst] as Item.User).isChecked = isChecked
        }
        if (indexOfLast > -1) {
            (items[indexOfLast] as Item.User).isChecked = isChecked
        }
        notifyItemChanged(indexOfFirst)
        notifyItemChanged(indexOfLast)
    }

    class HeaderViewHolder(
        private val binding: KakaoSdkHeaderFriendPickerBinding
    ) : RecyclerView.ViewHolder(binding.root) {
        fun bind(header: Item.Header) {
            binding.titleTv.text = header.title
        }
    }

    class DividerViewHolder(binding: KakaoSdkCategoryDividerBinding) :
        RecyclerView.ViewHolder(binding.root)

    class EmptySearchViewHolder(
        private val isPopup: Boolean,
        private val binding: KakaoSdkEmptySearchResultTextBinding
    ) : RecyclerView.ViewHolder(binding.root) {
        fun setLayout() {
            binding.apply {
                if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                    if (isPopup) {
                        emptySearchResultText.apply {
                            (layoutParams as LinearLayout.LayoutParams).topMargin =
                                resources.getDimensionPixelSize(R.dimen.popup_empty_search_result_margin_top)
                        }
                    }
                } else {
                    emptySearchResultText.apply {
                        (layoutParams as LinearLayout.LayoutParams).topMargin = if (isPopup) {
                            resources.getDimensionPixelSize(R.dimen.popup_empty_search_result_margin_top_landscape)
                        } else {
                            resources.getDimensionPixelSize(R.dimen.empty_search_result_margin_top_landscape)
                        }
                    }
                }
            }
        }
    }

    class UserViewHolder(
        private val isPopup: Boolean,
        private val binding: KakaoSdkItemFriendPickerBinding,
        private val checkCallback: (Item.User) -> Unit,
        private val uncheckCallback: (Item.User) -> Unit
    ) : RecyclerView.ViewHolder(binding.root) {

        fun bind(user: Item.User, isSingle: Boolean) {
            binding.apply {
                if (user.disableSelectionReason == null) {
                    // 수신 거부 상태가 아닐 때
                    checkBox.visibility = View.VISIBLE
                    setAlpha(1f)
                    root.setOnClickListener {
                        if (!checkBox.isChecked) {
                            checkCallback(user)
                        } else {
                            uncheckCallback(user)
                        }
                    }
                } else {
                    // 수신 거부 상태일 때
                    checkBox.visibility = View.GONE
                    setAlpha(0.3f)
                    root.setOnClickListener(null) // 선택 불가능하도록 처리
                }

                setProfileImageView(user)
                setNicknameTextView(isPopup, user)
                setRejectTextView(user)
                setCheckbox(isSingle, user)
            }
        }

        private fun setCheckbox(isSingle: Boolean, user: Item.User) {
            binding.checkBox.apply {
                val checkBoxDrawable = if (isSingle) {
                    R.drawable.kakao_sdk_daynight_radio
                } else {
                    R.drawable.kakao_sdk_daynight_checkbox
                }

                setButtonDrawable(checkBoxDrawable)
                isChecked = user.isChecked
            }
        }

        private fun setProfileImageView(user: Item.User) {
            binding.apply {
                val defaultProfileImage =
                    getDefaultProfileImage(user.uuid.hashCode(), root.context).resourceId
                ImageLoader.get(itemView.context).load(user.profileImageUrl)
                    .placeHolder(defaultProfileImage).into(profileIv)
            }
        }

        private fun setNicknameTextView(isPopup: Boolean, user: Item.User) {
            binding.nicknameTv.apply {
                val topMargin = if (isPopup) {
                    if (user.disableSelectionReason == null) resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_margin_top)
                    else resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_reject_margin_top)
                } else {
                    if (user.disableSelectionReason == null) resources.getDimensionPixelSize(R.dimen.list_item_nickname_margin_top)
                    else resources.getDimensionPixelSize(R.dimen.list_item_nickname_reject_margin_top)
                }
                val bottomMargin = if (isPopup) {
                    if (user.disableSelectionReason == null) resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_margin_bottom)
                    else resources.getDimensionPixelSize(R.dimen.popup_list_item_nickname_reject_margin_bottom)
                } else {
                    if (user.disableSelectionReason == null) resources.getDimensionPixelSize(R.dimen.list_item_nickname_margin_bottom)
                    else resources.getDimensionPixelSize(R.dimen.list_item_nickname_reject_margin_bottom)
                }
                text = user.nickname
                ConstraintSet().apply {
                    clone(binding.root)
                    connect(id, TOP, PARENT_ID, TOP, topMargin)
                    connect(id, BOTTOM, PARENT_ID, BOTTOM, bottomMargin)
                }.applyTo(binding.root)
            }
        }

        private fun setRejectTextView(user: Item.User) {
            binding.rejectTv.visibility =
                if (user.disableSelectionReason == null) View.GONE else View.VISIBLE
        }

        private fun setAlpha(value: Float) {
            binding.apply {
                profileIv.alpha = value
                nicknameTv.alpha = value
                rejectTv.alpha = value
            }
        }
    }

    companion object {
        const val VIEW_TYPE_HEADER = 0
        const val VIEW_TYPE_USER = 1
        const val VIEW_CATEGORY_DIVIDER = 2
        const val VIEW_EMPTY_SEARCH_RESULT_TEXT = 3
    }
}

internal class ItemDiffUtilCallback : DiffUtil.Callback() {
    private lateinit var oldList: List<Item>
    private lateinit var newList: List<Item>

    override fun getOldListSize(): Int = oldList.size
    override fun getNewListSize(): Int = newList.size

    fun setLists(oldList: List<Item>, newList: List<Item>) {
        this.oldList = oldList
        this.newList = newList
    }

    override fun areItemsTheSame(oldItemPosition: Int, newItemPosition: Int): Boolean {
        return oldList[oldItemPosition].hashCode() == newList[newItemPosition].hashCode()
    }

    override fun areContentsTheSame(oldItemPosition: Int, newItemPosition: Int): Boolean {
        val oldItem = oldList[oldItemPosition]
        val newItem = newList[newItemPosition]
        when (oldItem) {
            is Item.Header -> {
                if (newItem !is Item.Header) {
                    return false
                }
                return oldItem.title == newItem.title
            }
            is Item.User -> {
                if (newItem !is Item.User) {
                    return false
                }
                return oldItem.uuid == newItem.uuid
            }
            is Item.Divider -> {
                if (newItem !is Item.Divider) {
                    return false
                }
                return true
            }
            is Item.EmptySearchResult -> {
                if (newItem !is Item.EmptySearchResult) {
                    return false
                }
                return true
            }
        }
    }
}